import datetime
import functools
import unittest

import numpy as np

from clouddrift.binning import (
    DEFAULT_BINS_NUMBER,
    DEFAULT_COORD_NAME,
    DEFAULT_DATA_NAME,
    _is_datetime_array,
    binned_statistics,
)


class binning_tests(unittest.TestCase):
    def setUp(self):
        self.coords_1d = np.array(
            [-0.2, 0.3, 0.6, 0.7, 1.2, 1.3, 1.8, 2.1, 2.7, 2.8, 3.1]
        )
        self.date_1d = np.datetime64("2020-01-01 00:00") + np.arange(
            11
        ) * np.timedelta64(1, "D")

        self.values_1d = np.array([1, 2, 3, 4, 5, 6, 9, 20, 20, 20, 20])
        self.bins_range_1d = (0, 3)

        self.coords_2d = [self.coords_1d, self.coords_1d]
        self.bins_range_2d = [(0, 3), (0, 3)]

        self.coords_3d = [self.coords_1d, self.coords_1d, self.coords_1d]
        self.bins_range_3d = [(0, 3), (0, 3), (0, 3)]

        # example with 8 regions around x=[-1,0,1], y=[-1,0,1], z=[-1,0,1]
        self.coords_3d_ex = [[], [], []]
        self.values_3d_ex = []
        for x in [-0.5, -0.25, 0.25, 0.5]:
            for y in [-0.5, -0.25, 0.25, 0.5]:
                for z in [-0.5, -0.25, 0.25, 0.5]:
                    self.coords_3d_ex[0].append(x)
                    self.coords_3d_ex[1].append(y)
                    self.coords_3d_ex[2].append(z)
                    self.values_3d_ex.append(int(x > 0) + int(y > 0) + int(z > 0))

    def test_parameters_dimensions(self):
        with self.assertRaises(ValueError):
            binned_statistics(self.coords_1d, dim_names=["x", "y"])

        with self.assertRaises(ValueError):
            binned_statistics(self.coords_1d, bins=[10, 10])

        with self.assertRaises(ValueError):
            binned_statistics(self.coords_2d, bins=[10])

        with self.assertRaises(ValueError):
            binned_statistics(self.coords_2d, dim_names=["x"])

        with self.assertRaises(ValueError):
            binned_statistics(self.coords_1d, output_names=["x", "y"])

        with self.assertRaises(ValueError):
            binned_statistics(
                self.coords_1d,
                data=np.ones_like(self.coords_1d),
                output_names=["x", "y"],
            )
        with self.assertRaises(ValueError):
            binned_statistics(self.coords_1d, data=np.ones((len(self.coords_1d), 1)))

    def test_is_datetime(self):
        arr = np.array([np.datetime64("2020-01-01"), np.datetime64("2020-01-02")])
        self.assertTrue(_is_datetime_array(arr))

        arr = np.array(
            [np.datetime64("2020-01-01"), np.datetime64("2020-01-02")], dtype="O"
        )
        self.assertTrue(_is_datetime_array(arr))

        arr = np.array([datetime.date(2020, 1, 1), datetime.date(2020, 1, 2)])
        self.assertTrue(_is_datetime_array(arr))

        arr = np.array([1, 2, 3])
        self.assertFalse(_is_datetime_array(arr))

        arr = np.array([1, 2, 3], dtype="O")
        self.assertFalse(_is_datetime_array(arr))

        arr = np.array([None], dtype="O")
        self.assertFalse(_is_datetime_array(arr))

        arr = np.array([np.nan, np.nan])
        self.assertFalse(_is_datetime_array(arr))

    def test_bins_number_default(self):
        ds = binned_statistics(self.coords_1d)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), DEFAULT_BINS_NUMBER)

        ds = binned_statistics(self.coords_2d)
        for v in ds.sizes.values():
            self.assertEqual(v, DEFAULT_BINS_NUMBER)

        ds = binned_statistics(self.coords_3d, bins=5)
        for v in ds.sizes.values():
            self.assertEqual(v, 5)

        ds = binned_statistics(self.coords_2d, bins=(5, None))
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), 5)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_1"]), DEFAULT_BINS_NUMBER)

    def test_bins_list(self):
        ds = binned_statistics(self.coords_1d, bins=[[0, 1, 2, 3]])
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), 3)

        ds = binned_statistics(self.coords_1d, bins=[np.arange(0, 4, 1)])
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), 3)

        ds = binned_statistics(self.coords_1d, bins=[np.arange(0, 4, 0.5)])
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), 7)

    def test_1d_hist_number(self):
        ds = binned_statistics(self.coords_1d, bins=3)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), 3)

    def test_2d_hist_number(self):
        ds = binned_statistics(self.coords_2d, bins=(3, 4))
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), 3)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_1"]), 4)

    def test_3d_hist_number(self):
        ds = binned_statistics(self.coords_3d, bins=(3, 4, 5))
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), 3)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_1"]), 4)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_2"]), 5)

    def test_hist_center(self):
        for i in range(1, 10):
            ds = binned_statistics(self.coords_1d, bins=i)

            bins_coords = np.linspace(
                np.min(self.coords_1d), np.max(self.coords_1d), i + 1
            )
            bins_center = (bins_coords[:-1] + bins_coords[1:]) / 2

            self.assertIsNone(
                np.testing.assert_allclose(
                    ds[f"{DEFAULT_COORD_NAME}_0"].values, bins_center
                )
            )

    def test_1d_output(self):
        ds = binned_statistics(coords=self.coords_1d)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), DEFAULT_BINS_NUMBER)
        self.assertEqual(
            sum(ds[f"{DEFAULT_DATA_NAME}_count"].values), len(self.coords_1d)
        )

    def test_1d_output_bins(self):
        n_bins = 3
        ds = binned_statistics(coords=self.coords_1d, bins=n_bins)
        self.assertEqual(len(ds[f"{DEFAULT_COORD_NAME}_0"]), n_bins)
        self.assertEqual(
            sum(ds[f"{DEFAULT_DATA_NAME}_count"].values), len(self.coords_1d)
        )
        self.assertEqual(len(ds[f"{DEFAULT_DATA_NAME}_count"].shape), 1)

    def test_1d_output_variables_mean(self):
        value_1 = 1
        value_2 = 2
        variable = [
            np.ones_like(self.coords_1d) * value_1,
            np.ones_like(self.coords_1d) * value_2,
        ]
        ds = binned_statistics(
            coords=self.coords_1d, data=variable, bins=2, statistics="mean"
        )
        self.assertEqual(len(ds.data_vars), 2)
        self.assertTrue(all(ds[f"{DEFAULT_DATA_NAME}_0_mean"] == value_1))
        self.assertTrue(all(ds[f"{DEFAULT_DATA_NAME}_1_mean"] == value_2))

        coords_threshold = 1
        variables = np.ones_like(self.coords_1d) * value_1
        variables[self.coords_1d > coords_threshold] = value_2
        ds = binned_statistics(
            coords=self.coords_1d,
            data=variables,
            bins=3,
            bins_range=(0, 3),
            statistics="mean",
        )
        mask = ds[f"{DEFAULT_COORD_NAME}_0"].values < coords_threshold
        self.assertTrue(all(ds[f"{DEFAULT_DATA_NAME}_0_mean"].values[mask] == value_1))
        self.assertTrue(all(ds[f"{DEFAULT_DATA_NAME}_0_mean"].values[~mask] == value_2))

    def test_2d_output(self):
        ds = binned_statistics(coords=self.coords_2d)
        for v in ds.sizes.values():
            self.assertEqual(v, DEFAULT_BINS_NUMBER)
        self.assertEqual(len(ds[f"{DEFAULT_DATA_NAME}_count"].shape), 2)

        n_bins = (3, 4)
        ds = binned_statistics(coords=self.coords_2d, bins=n_bins)
        for i, v in enumerate(ds.sizes.values()):
            self.assertEqual(v, n_bins[i])
        self.assertEqual(len(ds[f"{DEFAULT_DATA_NAME}_count"].shape), 2)

        n_bins = (3, None)
        ds = binned_statistics(coords=self.coords_2d, bins=n_bins)
        for i, v in enumerate(ds.sizes.values()):
            self.assertEqual(
                v, n_bins[i] if n_bins[i] is not None else DEFAULT_BINS_NUMBER
            )
        self.assertEqual(len(ds[f"{DEFAULT_DATA_NAME}_count"].shape), 2)

    def test_3d_output(self):
        ds = binned_statistics(coords=self.coords_3d)
        for v in ds.sizes.values():
            self.assertEqual(v, DEFAULT_BINS_NUMBER)
        self.assertEqual(len(ds[f"{DEFAULT_DATA_NAME}_count"].shape), 3)

        n_bins = (3, 4, 5)
        ds = binned_statistics(coords=self.coords_3d, bins=n_bins)
        for i, v in enumerate(ds.sizes.values()):
            self.assertEqual(v, n_bins[i])
        self.assertEqual(len(ds[f"{DEFAULT_DATA_NAME}_count"].shape), 3)

        n_bins = (3, 4, None)
        ds = binned_statistics(coords=self.coords_3d, bins=n_bins)
        for i, v in enumerate(ds.sizes.values()):
            self.assertEqual(
                v, n_bins[i] if n_bins[i] is not None else DEFAULT_BINS_NUMBER
            )
        self.assertEqual(len(ds[f"{DEFAULT_DATA_NAME}_count"].shape), 3)

    def test_3d_output_mean_example(self):
        ds = binned_statistics(
            coords=self.coords_3d_ex,
            data=self.values_3d_ex,
            bins=(2, 2, 2),
            bins_range=[(-1, 1), (-1, 1), (-1, 1)],
            statistics="mean",
        )
        self.assertEqual(len(ds.data_vars), 1)
        self.assertIsNone(
            np.testing.assert_allclose(
                ds[f"{DEFAULT_DATA_NAME}_0_mean"].values.flatten(),
                np.array(
                    [
                        0,  # (-1, -1, -1)
                        1,  # (-1, -1, 0)
                        1,  # (-1, 0, -1)
                        2,  # (-1, 0, 0)
                        1,  # (0, -1, -1)
                        2,  # (0, -1, 0)
                        2,  # (0, 0, -1)
                        3,  # (0, 0, 0)
                    ]
                ),
            )
        )

    def test_hist_range(self):
        ds = binned_statistics(
            coords=self.coords_1d, bins=3, bins_range=self.bins_range_1d
        )
        self.assertEqual(
            sum(ds[f"{DEFAULT_DATA_NAME}_count"].values),
            len(
                self.coords_1d[
                    np.logical_and(
                        self.coords_1d >= self.bins_range_1d[0],
                        self.coords_1d < self.bins_range_1d[1],
                    )
                ]
            ),
        )

    def test_hist_range_2d(self):
        ds = binned_statistics(
            coords=self.coords_2d, bins=(3, 3), bins_range=self.bins_range_2d
        )
        self.assertEqual(
            sum(ds[f"{DEFAULT_DATA_NAME}_count"].values.flatten()),
            len(
                self.coords_1d[
                    np.logical_and(
                        self.coords_1d >= self.bins_range_1d[0],
                        self.coords_1d < self.bins_range_1d[1],
                    )
                ]
            ),
        )

    def test_hist_range_3d(self):
        ds = binned_statistics(
            coords=self.coords_3d, bins=(3, 3, 3), bins_range=self.bins_range_3d
        )
        self.assertEqual(
            sum(ds[f"{DEFAULT_DATA_NAME}_count"].values.flatten()),
            len(
                self.coords_1d[
                    np.logical_and(
                        self.coords_1d >= self.bins_range_1d[0],
                        self.coords_1d < self.bins_range_1d[1],
                    )
                ]
            ),
        )

    def test_rename_dimensions(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            bins=4,
            dim_names=["x"],
        )
        self.assertIn("x", ds.sizes)

        ds = binned_statistics(
            coords=self.coords_2d,
            bins=(3, 4),
            dim_names=["x", "y"],
        )
        self.assertIn("x", ds.sizes)
        self.assertIn("y", ds.sizes)

    def test_coords_finite(self):
        coords_inf = self.coords_1d.copy()
        coords_inf[0] = np.inf

        with self.assertRaises(ValueError):
            binned_statistics(
                coords=coords_inf,
                bins=3,
                data=self.values_1d,
            )

        coords_nan = [self.coords_1d.copy(), self.coords_1d.copy()]
        coords_nan[1][0] = np.nan
        with self.assertRaises(ValueError):
            binned_statistics(
                coords=coords_nan,
                bins=3,
                data=self.values_1d,
            )

        coords_nan = [self.date_1d.copy()]
        coords_nan[0][0] = np.datetime64("NaT")
        with self.assertRaises(ValueError):
            binned_statistics(
                coords=coords_nan,
                bins=3,
                data=self.values_1d,
            )

    def test_rename_variables(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            bins=4,
            output_names=["x"],
        )
        self.assertIn("x_count", ds.data_vars)

        ds = binned_statistics(
            coords=self.coords_2d,
            data=[self.coords_1d, self.coords_1d],
            output_names=["x", "y"],
        )
        self.assertIn("x_count", ds.data_vars)
        self.assertIn("y_count", ds.data_vars)

        ds = binned_statistics(
            coords=self.coords_2d,
            data=[self.coords_1d, self.coords_1d],
            output_names=["x", "y"],
            statistics=["count", "mean"],
        )
        self.assertIn("x_count", ds.data_vars)
        self.assertIn("y_count", ds.data_vars)
        self.assertIn("x_mean", ds.data_vars)
        self.assertIn("y_mean", ds.data_vars)

    def test_statistics_wrong_function(self):
        with self.assertRaises(ValueError):
            binned_statistics(
                coords=self.coords_1d,
                data=self.values_1d,
                bins=3,
                statistics=["non_existent_function"],
            )

    def test_statistics_wrong_type(self):
        with self.assertRaises(ValueError):
            binned_statistics(
                coords=self.coords_1d,
                data=self.values_1d,
                bins=3,
                statistics=["mean", 42],
            )

        with self.assertRaises(ValueError):
            binned_statistics(
                coords=self.coords_1d,
                data=self.values_1d,
                bins=3,
                statistics=np.array([1, 2, 3]),
            )

    def test_statistics_default(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_count", ds.data_vars)

    def test_statistics_no_precalculated_values(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
            statistics="mean",
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean", ds.data_vars)

        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
            statistics="std",
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_std", ds.data_vars)

    def test_statistics_all(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
            statistics=["count", "sum", "mean", "std", "min", "max"],
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_count", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_sum", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_std", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_min", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_max", ds.data_vars)

    def test_statistics_n_vars(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=[self.values_1d, self.values_1d, self.values_1d],
            statistics=["count", "sum", "mean", "std", "min", "max"],
        )
        for var in [f"{DEFAULT_DATA_NAME}_0", f"{DEFAULT_DATA_NAME}_1"]:
            self.assertIn(f"{var}_count", ds.data_vars)
            self.assertIn(f"{var}_sum", ds.data_vars)
            self.assertIn(f"{var}_mean", ds.data_vars)
            self.assertIn(f"{var}_std", ds.data_vars)
            self.assertIn(f"{var}_min", ds.data_vars)
            self.assertIn(f"{var}_max", ds.data_vars)

    def test_statistics_n_vars_rename(self):
        var_names = ["u", "v", "w"]
        ds = binned_statistics(
            coords=self.coords_1d,
            data=[self.values_1d, self.values_1d, self.values_1d],
            statistics=["count", "sum", "mean", "median", "std", "min", "max"],
            output_names=var_names,
        )
        for var in var_names:
            self.assertIn(f"{var}_count", ds.data_vars)
            self.assertIn(f"{var}_sum", ds.data_vars)
            self.assertIn(f"{var}_mean", ds.data_vars)
            self.assertIn(f"{var}_median", ds.data_vars)
            self.assertIn(f"{var}_std", ds.data_vars)
            self.assertIn(f"{var}_min", ds.data_vars)
            self.assertIn(f"{var}_max", ds.data_vars)

    def test_statistics_callable(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
            statistics=[np.median, "mean"],
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_median", ds.data_vars)

        np.testing.assert_allclose(
            ds[f"{DEFAULT_DATA_NAME}_0_mean"].values,
            np.array([2.5, 6.666, 20.0]),
            rtol=1e-2,
        )

        np.testing.assert_allclose(
            ds[f"{DEFAULT_DATA_NAME}_0_median"].values,
            np.array([2.5, 6, 20.0]),
            rtol=1e-2,
        )

    def test_statistics_callable_partial(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
            statistics=[
                functools.partial(np.percentile, q=25),
                functools.partial(np.percentile, q=50),
                functools.partial(np.percentile, q=75),
                "mean",
            ],
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_percentile", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_percentile_1", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_percentile_2", ds.data_vars)

    def test_statistics_callable_lambda(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
            statistics=[lambda x: np.percentile(x, q=40), "mean"],
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_stat", ds.data_vars)

    def test_statistics_callable_multiple_lambda(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=self.values_1d,
            bins=3,
            statistics=[
                lambda x: np.percentile(x, q=10),
                lambda x: np.percentile(x, q=20),
                lambda x: np.percentile(x, q=30),
                np.mean,
                np.mean,
            ],
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_stat", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_stat_1", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_stat_2", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean_1", ds.data_vars)

    def test_statistics_multivariable_wrong_parameter(self):
        with self.assertRaises(ValueError):
            binned_statistics(
                coords=self.coords_1d,
                data=[self.values_1d, self.values_1d],
                bins=3,
                statistics=(
                    0,
                    lambda data: np.sqrt(np.mean(data[0] ** 2 + data[1] ** 2)),
                ),
            )

        with self.assertRaises(ValueError):
            binned_statistics(
                coords=self.coords_1d,
                data=[self.values_1d, self.values_1d],
                bins=3,
                statistics=(
                    "ke",
                    0,
                ),
            )

    def test_statistics_multivariable(self):
        ds = binned_statistics(
            coords=self.coords_1d,
            data=[np.ones_like(self.values_1d) * 3, np.ones_like(self.values_1d) * 4],
            bins=3,
            statistics=[
                "mean",
                "count",
                (
                    "ke",
                    lambda data: np.sqrt(np.mean(data[0] ** 2 + data[1] ** 2)),
                ),
            ],
        )
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_count", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_1_count", ds.data_vars)
        self.assertIn(f"{DEFAULT_DATA_NAME}_0_mean", ds.data_vars)
        self.assertTrue(all(ds[f"{DEFAULT_DATA_NAME}_0_mean"] == 3))
        self.assertIn(f"{DEFAULT_DATA_NAME}_1_mean", ds.data_vars)
        self.assertTrue(all(ds[f"{DEFAULT_DATA_NAME}_1_mean"] == 4))
        self.assertIn("ke", ds.data_vars)
        self.assertTrue(all(ds["ke"] == 5))

    def test_statistics_complex_mean_sum_count_std(self):
        # Complex input for mean, sum, count, std
        coords = self.coords_1d
        values = np.array(
            [
                1 + 1j,
                2 + 2j,
                3 + 3j,
                4 + 4j,
                5 + 5j,
                6 + 6j,
                7 + 7j,
                8 + 8j,
                9 + 9j,
                10 + 10j,
                11 + 11j,
            ]
        )
        ds = binned_statistics(
            coords=coords,
            data=values,
            bins=3,
            statistics=["mean", "sum", "count", "std"],
        )
        # Check that the output is complex for mean, sum, std
        self.assertTrue(np.iscomplexobj(ds[f"{DEFAULT_DATA_NAME}_0_mean"].values))
        self.assertTrue(np.iscomplexobj(ds[f"{DEFAULT_DATA_NAME}_0_sum"].values))
        self.assertFalse(np.iscomplexobj(ds[f"{DEFAULT_DATA_NAME}_0_std"].values))
        self.assertTrue(
            np.issubdtype(ds[f"{DEFAULT_DATA_NAME}_0_std"].dtype, np.floating)
        )
        # Count should be real and integer
        self.assertTrue(
            np.issubdtype(ds[f"{DEFAULT_DATA_NAME}_0_count"].dtype, np.integer)
        )
        # Check that the sum is the sum of values in each bin
        self.assertAlmostEqual(
            np.sum(ds[f"{DEFAULT_DATA_NAME}_0_sum"].values), np.sum(values)
        )
        # Check that the mean is the mean of values in each bin
        mask0 = (coords >= -0.2) & (coords < 0.9)
        mask1 = (coords >= 0.9) & (coords < 2.0)
        mask2 = (coords >= 2.0) & (coords <= 3.1)
        means = [
            np.mean(values[mask0]) if np.any(mask0) else 0,
            np.mean(values[mask1]) if np.any(mask1) else 0,
            np.mean(values[mask2]) if np.any(mask2) else 0,
        ]
        np.testing.assert_allclose(
            ds[f"{DEFAULT_DATA_NAME}_0_mean"].values, means, rtol=1e-12
        )

    def test_statistics_complex_min_max_median_raises(self):
        coords = self.coords_1d
        values = np.array(
            [
                1 + 1j,
                2 + 2j,
                3 + 3j,
                4 + 4j,
                5 + 5j,
                6 + 6j,
                7 + 7j,
                8 + 8j,
                9 + 9j,
                10 + 10j,
                11 + 11j,
            ]
        )
        with self.assertRaises(ValueError):
            binned_statistics(
                coords=coords,
                data=values,
                bins=3,
                statistics="min",
            )

        with self.assertRaises(ValueError):
            binned_statistics(
                coords=coords,
                data=values,
                bins=3,
                statistics="max",
            )

        with self.assertRaises(ValueError):
            binned_statistics(
                coords=coords,
                data=values,
                bins=3,
                statistics="median",
            )

    def test_statistics_datetime_coords(self):
        coords = [self.coords_1d, self.date_1d]

        ds = binned_statistics(
            coords=coords,
            data=self.values_1d,
            bins=3,
            dim_names=["x", "time"],
            statistics=["count", "mean"],
        )

        # check if datetime data is handled correctly
        assert ds["time"].dtype.kind == "M"

    def test_statistics_datetime_data_sum(self):
        coords = [self.coords_1d, self.date_1d]
        data = [self.values_1d, self.date_1d]

        with self.assertRaises(ValueError):
            binned_statistics(
                coords=coords,
                data=data,
                bins=3,
                dim_names=["x", "time"],
                statistics=["sum"],  # not supported for datetime data
            )

    def test_statistics_datetime_data(self):
        coords = [self.coords_1d, self.date_1d]
        data = [self.values_1d, self.date_1d]

        ds = binned_statistics(
            coords=coords,
            data=data,
            bins=3,
            dim_names=["x", "time"],
            statistics=["count", "mean", "median", "std", "min", "max"],
        )

        # check if datetime data is handled correctly
        assert ds["time"].dtype.kind == "M"

    def test_statistics_datetime_data_values(self):
        coords = [self.date_1d]
        data = [self.date_1d]

        ds = binned_statistics(
            coords=coords,
            data=data,
            bins=3,
            dim_names=["time"],
            statistics=["count", "mean", "median", "std", "min", "max"],
        )

        # Check that the output is date type
        for var in ds.data_vars:
            if not (var.endswith("count") or var.endswith("std")):
                self.assertTrue(ds[var].dtype.kind == "M")
        self.assertTrue(ds[f"{DEFAULT_DATA_NAME}_0_std"].dtype.kind == "m")

        self.assertIsNone(
            np.testing.assert_array_equal(ds[f"{DEFAULT_DATA_NAME}_0_count"], [4, 3, 4])
        )

        # create mask for three bins
        mask0 = self.date_1d <= self.date_1d[3]
        mask1 = (self.date_1d > self.date_1d[3]) & (self.date_1d < self.date_1d[7])
        mask2 = self.date_1d >= self.date_1d[7]

        self.assertIsNone(
            np.testing.assert_array_equal(
                ds[f"{DEFAULT_DATA_NAME}_0_mean"].values,
                np.array(
                    [
                        "2020-01-02T12:00:00",
                        "2020-01-06T00:00:00",
                        "2020-01-09T12:00:00",
                    ],
                    dtype="datetime64[s]",
                ),
            )
        )

        self.assertIsNone(
            np.testing.assert_array_equal(
                ds[f"{DEFAULT_DATA_NAME}_0_median"].values,
                np.array(
                    [
                        "2020-01-02T12:00:00",
                        "2020-01-06T00:00:00",
                        "2020-01-09T12:00:00",
                    ],
                    dtype="datetime64[s]",
                ),
            )
        )

        self.assertIsNone(
            np.testing.assert_array_equal(
                ds[f"{DEFAULT_DATA_NAME}_0_std"].values,
                np.array(
                    [96598, 70545, 96598],
                    dtype="timedelta64[s]",
                ),
            )
        )

        self.assertIsNone(
            np.testing.assert_array_equal(
                ds[f"{DEFAULT_DATA_NAME}_0_min"].values,
                [
                    min(self.date_1d[mask0]),
                    min(self.date_1d[mask1]),
                    min(self.date_1d[mask2]),
                ],
            )
        )

        self.assertIsNone(
            np.testing.assert_array_equal(
                ds[f"{DEFAULT_DATA_NAME}_0_max"].values,
                [
                    max(self.date_1d[mask0]),
                    max(self.date_1d[mask1]),
                    max(self.date_1d[mask2]),
                ],
            )
        )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys

import clouddrift

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------

project = "CloudDrift"
copyright = "2022-2023, CloudDrift"
author = "Philippe Miron"
version = clouddrift.version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx_copybutton",
]


autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_book_theme"  # alabaster, sphinx_rtd_theme

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path: list[str] = []
html_logo = "logo.png"
html_favicon = "favicon.ico"
html_theme_options = {
    "logo": {
        "image_light": "logo.png",
        "image_dark": "logo.png",
    }
}

# sphinx-copybutton settings
copybutton_exclude = ".linenos, .gp"  # don't copy prompts or line numbers



"""Module for binning Lagrangian data."""

import datetime
import warnings
from functools import partial, wraps
from typing import Callable

import numpy as np
import pandas as pd
import xarray as xr

DEFAULT_BINS_NUMBER = 10
DEFAULT_COORD_NAME = "coord"
DEFAULT_DATA_NAME = "data"


def binned_statistics(
    coords: np.ndarray | list[np.ndarray],
    data: np.ndarray | list[np.ndarray] | None = None,
    bins: int | list = DEFAULT_BINS_NUMBER,
    bins_range: list | None = None,
    dim_names: list[str] | None = None,
    output_names: list[str] | None = None,
    statistics: str | list | Callable[[np.ndarray], float] = "count",
) -> xr.Dataset:
    """
    Perform N-dimensional binning and compute statistics of values in each bin. The result is returned as an Xarray Dataset.

    Parameters
    ----------
    coords : array-like or list of array-like
        Array(s) of Lagrangian data coordinates to be binned. For 1D, provide a single array.
        For N-dimensions, provide a list of N arrays, each giving coordinates along one dimension.
    data : array-like or list of array-like
        Data values associated with the Lagrangian coordinates in coords.
        Can be a single array or a list of arrays for multiple variables.
        Complex values are supported for the supported statistics except for 'min', 'max', and 'median'.
    bins : int or lists, optional
        Number of bins or bin edges per dimension. It can be:
        - An int: same number of bins for all dimensions,
        - A list of ints or arrays: one per dimension, specifying either bin count or bin edges,
        - None: defaults to 10 bins per dimension.
    bins_range : list of tuples, optional
        Outer bin limits for each dimension.
    statistics : str or list of str, Callable[[np.ndarray], float] or list[Callable[[np.ndarray], float]]
        Statistics to compute for each bin. It can be:
        - a string, supported values: 'count', 'sum', 'mean', 'median', 'std', 'min', 'max', (default: "count"),
        - a custom function as a callable for univariate statistics that take a 1D array of values and return a single value.
          The callable is applied to each variable of data.
        - a tuple of (output_name, callable) for multivariate statistics. 'output_name' is used to identify the resulting variables.
          In this case, the callable will receive the list of arrays provided in `data`. For example, to calculate kinetic energy from data with velocity components `u` and `v`,
          you can pass `data = [u, v]` and  `statistics=("ke", lambda data: np.sqrt(np.mean(data[0] ** 2 + data[1] ** 2)))`.
        - a list containing any combination of the above, e.g., ['mean', np.nanmax, ('ke', lambda data: np.sqrt(np.mean(data[0] ** 2 + data[1] ** 2)))].
    dim_names : list of str, optional
        Names for the dimensions of the output xr.Dataset.
        If None, default names are "coord_0", "coord_1", etc.
    output_names : list of str, optional
        Names for output variables in the xr.Dataset.
        If None, default names are "data_0_{statistic}", "data_1_{statistic}", etc.

    Returns
    -------
    xr.Dataset
        Xarray dataset with binned means and count for each variable.
    """
    # convert coords, data parameters to numpy arrays and validate dimensions
    # D, N = number of dimensions and number of data points
    if not isinstance(coords, np.ndarray) or coords.ndim == 1:
        coords = np.atleast_2d(coords)
    D, N = coords.shape

    # validate coordinates are finite
    for c in coords:
        var = c.copy()
        if var.dtype == "O":
            var = var.astype(type(var[0]))
        if _is_datetime_array(var):
            if pd.isna(var).any():
                raise ValueError("Datetime coordinates must be finite values.")
        else:
            if pd.isna(var).any() or np.isinf(var).any():
                raise ValueError("Coordinates must be finite values.")

    # V, VN = number of variables and number of data points per variable
    if data is None:
        data = np.empty((1, 0))
        V, VN = 1, N  # no data provided
    elif not isinstance(data, np.ndarray) or data.ndim == 1:
        data = np.atleast_2d(data)
        V, VN = data.shape
    else:
        V, VN = data.shape

    # convert datetime coordinates to numeric values
    coords_datetime_index = np.where([_is_datetime_array(c) for c in coords])[0]
    for i in coords_datetime_index:
        coords[i] = _datetime64_to_float(coords[i])
    coords = coords.astype(np.float64)

    # set default bins and bins range
    if isinstance(bins, (list, tuple)):
        if len(bins) != len(coords):
            raise ValueError("`bins` must match the dimensions of the coordinates")
        bins = [b if b is not None else DEFAULT_BINS_NUMBER for b in bins]
    elif isinstance(bins, int):
        bins = [bins if bins is not None else DEFAULT_BINS_NUMBER] * len(coords)

    if bins_range is None:
        bins_range = [(np.nanmin(c), np.nanmax(c)) for c in coords]
    else:
        if isinstance(bins_range, tuple):
            bins_range = [bins_range] * len(coords)
        bins_range = [
            r if r is not None else (np.nanmin(c), np.nanmax(c))
            for r, c in zip(bins_range, coords)
        ]

    # validate statistics parameter
    ordered_statistics = ["count", "sum", "mean", "median", "std", "min", "max"]
    if isinstance(statistics, (str, tuple)) or callable(statistics):
        statistics = [statistics]
    elif not isinstance(statistics, list):
        raise ValueError(
            "`statistics` must be a string, list of strings, Callable, or a list of Callables. "
            f"Supported values: {', '.join(ordered_statistics)}."
        )
    if invalid := [
        stat
        for stat in statistics
        if (stat not in ordered_statistics)
        and not callable(stat)
        and not isinstance(stat, tuple)
    ]:
        raise ValueError(
            f"Unsupported statistic(s): {', '.join(map(str, invalid))}. "
            f"Supported: {ordered_statistics} or a Callable."
        )

    # validate multivariable statistics
    for statistic in statistics:
        if isinstance(statistic, tuple):
            output_name, statistic = statistic
            if not isinstance(output_name, str):
                raise ValueError(
                    f"Invalid output name '{output_name}', must be a string."
                )
            if not callable(statistic):
                raise ValueError(
                    "Multivariable `statistics` function is not Callable, must provide as a tuple(output_name, Callable)."
                )

    # validate and sort statistics for efficiency
    statistics_str = [s for s in statistics if isinstance(s, str)]
    statistics_func = [s for s in statistics if not isinstance(s, str)]
    statistics = (
        sorted(
            set(statistics_str),
            key=lambda x: ordered_statistics.index(x),
        )
        + statistics_func
    )

    if statistics and not data.size:
        warnings.warn(
            f"no `data` provided, `statistics` ({statistics}) will be computed on the coordinates."
        )

    # set default dimension names
    if dim_names is None:
        dim_names = [f"{DEFAULT_COORD_NAME}_{i}" for i in range(len(coords))]
    else:
        dim_names = [
            name if name is not None else f"{DEFAULT_COORD_NAME}_{i}"
            for i, name in enumerate(dim_names)
        ]

    # set default variable names
    if output_names is None:
        output_names = [
            f"{DEFAULT_DATA_NAME}_{i}" if data[0].size else DEFAULT_DATA_NAME
            for i in range(len(data))
        ]
    else:
        output_names = [
            name if name is not None else f"{DEFAULT_DATA_NAME}_{i}"
            for i, name in enumerate(output_names)
        ]

    # ensure inputs are consistent
    if D != len(dim_names):
        raise ValueError("`coords` and `dim_names` must have the same length")
    if V != len(output_names):
        raise ValueError("`data` and `output_names` must have the same length")
    if N != VN:
        raise ValueError("`coords` and `data` must have the same number of data points")

    # edges and bin centers
    if isinstance(bins, int) or isinstance(bins[0], int):
        edges = [np.linspace(r[0], r[1], b + 1) for r, b in zip(bins_range, bins)]
    else:
        edges = [np.asarray(b) for b in bins]
    edges_sz = [len(e) - 1 for e in edges]
    n_bins = int(np.prod(edges_sz))
    bin_centers = [0.5 * (e[:-1] + e[1:]) for e in edges]

    # convert bin centers back to datetime64 for output dataset
    for i in coords_datetime_index:
        bin_centers[i] = _float_to_datetime64(bin_centers[i])

    # digitize coordinates into bin indices
    # modify edges to ensure the last edge is inclusive
    # by adding a small tolerance to the last edge (1s for date coordinates)
    edges_with_tol = [e.copy() for e in edges]
    for i, e in enumerate(edges_with_tol):
        e[-1] += np.finfo(float).eps if i not in coords_datetime_index else 1
    indices = [np.digitize(c, edges_with_tol[j]) - 1 for j, c in enumerate(coords)]
    valid = np.all(
        [(j >= 0) & (j < edges_sz[i]) for i, j in enumerate(indices)], axis=0
    )
    indices = [i[valid] for i in indices]

    # create an iterable of statistics to compute
    statistics_iter = []
    for statistic in statistics:
        if isinstance(statistic, str) or callable(statistic):
            for var, name in zip(data, output_names):
                statistics_iter.append((var, name, statistic))
        elif isinstance(statistic, tuple):
            output_name, statistic = statistic
            statistics_iter.append((data, output_name, statistic))

    ds = xr.Dataset()
    for var, name, statistic in statistics_iter:
        # count the number of points in each bin
        var_finite, indices_finite = _filter_valid_and_finite(var, indices, valid)
        flat_idx = np.ravel_multi_index(indices_finite, edges_sz)

        # convert object arrays to a common type
        if var_finite.dtype == "O":
            var_finite = var_finite.astype(type(var_finite[0]))

        # loop through statistics for the variable
        bin_count, bin_mean, bin_sum = None, None, None

        if statistic == "count":
            binned_stats = _binned_count(flat_idx, n_bins)
            bin_count = binned_stats.copy()
        elif statistic == "sum":
            if _is_datetime_array(var_finite):
                raise ValueError("Datetime data is not supported for 'sum' statistic.")
            binned_stats = _binned_sum(flat_idx, n_bins, values=var_finite)
            bin_sum = binned_stats.copy()
        elif statistic == "mean":
            binned_stats = _binned_mean(
                flat_idx,
                n_bins,
                values=var_finite,
                bin_counts=bin_count,
                bin_sum=bin_sum,
            )
            bin_mean = binned_stats.copy()
        elif statistic == "std":
            binned_stats = _binned_std(
                flat_idx,
                n_bins,
                values=var_finite,
                bin_counts=bin_count,
                bin_mean=bin_mean,
            )
        elif statistic == "min":
            binned_stats = _binned_min(
                flat_idx,
                n_bins,
                values=var_finite,
            )
        elif statistic == "max":
            binned_stats = _binned_max(
                flat_idx,
                n_bins,
                values=var_finite,
            )
        elif statistic == "median":
            if np.iscomplexobj(var_finite):
                raise ValueError(
                    "Complex values are not supported for 'median' statistic."
                )
            binned_stats = _binned_apply_func(
                flat_idx,
                n_bins,
                values=var_finite,
                func=np.median,
            )
        else:
            binned_stats = _binned_apply_func(
                flat_idx,
                n_bins,
                values=var_finite,
                func=statistic,
            )

        # add the binned statistics variable to the Dataset
        variable_name = (
            name
            if var_finite.ndim == 2
            else _get_variable_name(name, statistic, ds.data_vars)
            if callable(statistic)
            else f"{name}_{statistic}"
        )

        ds[variable_name] = xr.DataArray(
            binned_stats.reshape(edges_sz),
            dims=dim_names,
            coords=dict(zip(dim_names, bin_centers)),
        )

    return ds


def _get_variable_name(
    output_name: str,
    func: Callable,
    ds_vars: xr.core.dataset.DataVariables | dict[str, xr.DataArray],
) -> str:
    """
    Get the name of the function or a default name if it is a lambda function.

    Parameters
    ----------
    func : Callable
        Function to get the name of.
    output_name : str
        Name of the output variable to which the function is applied.
    ds_vars : dict[str, xr.DataArray]
        Dictionary of existing variables in the dataset to avoid name collisions.

    Returns
    -------
    str
        Name of the function or a custom function name for lambda function.
    """
    default_name = "stat"
    if isinstance(func, partial):
        function_name = getattr(func.func, "__name__", default_name)
    else:
        function_name = getattr(func, "__name__", default_name)
        if function_name == "<lambda>":
            function_name = default_name

    # avoid name collisions with existing variables
    # by adding a suffix if the name already exists
    base_name = f"{output_name}_{function_name}"
    name = base_name

    i = 1
    while name in ds_vars:
        name = f"{base_name}_{i}"
        i += 1

    return name


def _filter_valid_and_finite(
    var: np.ndarray, indices: list, valid: np.ndarray
) -> tuple[np.ndarray, list[np.ndarray]]:
    """
    Filter valid and finite values from the variable and indices.

    Args:
        var : np.ndarray
            Variable data to filter.
        indices : list
            List of index arrays to filter.
        valid : np.ndarray
            Boolean array indicating valid entries.
        V : int
            Size of the 'data' parameter to determine if the variable is multivariate.

    Returns:
        tuple[np.ndarray, list[np.ndarray]]: Filtered variable and indices.
    """
    if var.ndim == 2:
        var_valid = [v[valid] for v in var]
        mask = np.logical_or.reduce([~pd.isna(v) for v in var_valid])
        var_finite = np.array([v[mask] for v in var_valid])
        indices_finite = [i[mask] for i in indices]
    elif var.size:
        var = var[valid]
        mask = ~pd.isna(var)
        var_finite = var[mask]
        indices_finite = [i[mask] for i in indices]
    else:
        var_finite = var.copy()
        indices_finite = indices.copy()

    return var_finite, indices_finite


def _is_datetime_subelement(arr: np.ndarray) -> bool:
    """
    Get the type of the first non-null element in an array.

    Parameters
    ----------
    arr : np.ndarray
        Numpy array to check.

    Returns
    -------
    bool
        True if the first non-null element is a datetime type, False otherwise.
    """
    for item in arr.flat:
        if item is not None:
            return isinstance(item, (datetime.date, np.datetime64))
    return False


def _is_datetime_array(arr: np.ndarray) -> bool:
    """
    Verify if an array contains datetime values.

    Parameters
    ----------
    arr : np.ndarray
        Numpy array to check.

    Returns
    -------
    bool
        True if the array contains datetime64 or timedelta64 values, False otherwise.
    """
    if arr.dtype.kind == "M":  # numpy datetime64
        return True
    # if array is object, check first element
    if arr.dtype == object and arr.size > 0:
        return _is_datetime_subelement(arr)
    return False


def _datetime64_to_float(time_dt: np.ndarray) -> np.ndarray:
    """
    Convert np.datetime64 or array of datetime64 to float time since epoch.

    Parameters:
    ----------
    time_dt : np.datetime64 or array-like
        Datetime64 values to convert.

    Returns:
    -------
    float or np.ndarray of floats
        Seconds since UNIX epoch (1970-01-01T00:00:00).
    """
    reference_date = np.datetime64("1970-01-01T00:00:00")
    return np.array(
        (pd.to_datetime(time_dt) - pd.to_datetime(reference_date))
        / pd.to_timedelta(1, "s")
    )


def _float_to_datetime64(time_float, count=None):
    """
    Convert float seconds since UNIX epoch to np.datetime64.

    Parameters:
    ----------
    time_float : float or array-like
        Seconds since epoch (1970-01-01T00:00:00).

    Returns:
    -------
    np.datetime64 or np.ndarray of np.datetime64
        Converted datetime64 values.
    """
    reference_date = np.datetime64("1970-01-01T00:00:00")
    date = reference_date + time_float.astype("timedelta64[s]")

    return date


def handle_datetime_conversion(func: Callable) -> Callable:
    """
    A decorator to handle datetime64/timedelta64 conversion for
    statistics functions. For datetime `values`, it converts the time to float
    seconds since epoch before calling the function, and converts the result back
    to datetime64 after the function call.

    Assumes that the function accepts `values` as keyword arguments.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> np.ndarray:
        values = kwargs.get("values")

        datetime_conversion = False
        if values is not None:
            if datetime_conversion := _is_datetime_array(values):
                kwargs["values"] = _datetime64_to_float(values)

        # call back the original function
        result = func(*args, **kwargs)

        # Convert the result to datetime if necessary
        if datetime_conversion:
            if func.__name__ == "_binned_std":
                return result.astype("timedelta64[s]")
            return _float_to_datetime64(result)

        return result

    return wrapper


def _binned_count(flat_idx: np.ndarray, n_bins: int) -> np.ndarray:
    """
    Compute the count of values in each bin.

    Parameters
    ----------
    flat_idx : array-like
        1D array of bin indices, same shape as values.
    n_bins: int
        number of bins

    Returns
    -------
    result : array-like
        1D array of length n_bins with the count per bin
    """
    return np.bincount(flat_idx, minlength=n_bins)


@handle_datetime_conversion
def _binned_sum(flat_idx: np.ndarray, n_bins: int, values: np.ndarray) -> np.ndarray:
    """
    Compute the sum of values per bin.

    Parameters
    ----------
    flat_idx : array-like
        1D array of bin indices, same shape as values.
    n_bins: int
        number of bins
    values : array-like
        1D array of data values

    Returns
    -------
    result : array-like
        1D array of length n_bins with the sum per bin
    """
    if np.iscomplexobj(values):
        real = np.bincount(flat_idx, weights=values.real, minlength=n_bins)
        imag = np.bincount(flat_idx, weights=values.imag, minlength=n_bins)
        return real + 1j * imag
    else:
        return np.bincount(flat_idx, weights=values, minlength=n_bins)


@handle_datetime_conversion
def _binned_mean(
    flat_idx: np.ndarray,
    n_bins: int,
    values: np.ndarray,
    bin_counts: np.ndarray | None = None,
    bin_sum: np.ndarray | None = None,
) -> np.ndarray:
    """
    Compute the mean of values per bin.

    Parameters
    ----------
    flat_idx : array-like
        1D array of bin indices, same shape as values.
    n_bins: int
        number of bins
    values : array-like
        1D array of data values
    bin_counts : array-like, optional
        Precomputed counts per bin. If None, it will be computed using `_binned_count`.
    bin_sum : array-like, optional
        Precomputed sum per bin. If None, it will be computed using `_binned_sum`.

    Returns
    -------
    result : array-like
        1D array of length n_bins with the mean per bin
    """
    if bin_counts is None:
        bin_counts = _binned_count(flat_idx, n_bins)

    if bin_sum is None:
        bin_sum = _binned_sum(flat_idx, n_bins, values)

    return np.divide(
        bin_sum,
        bin_counts,
        out=np.full_like(bin_sum, np.nan, dtype=bin_sum.dtype),
        where=bin_counts > 0,
    )


@handle_datetime_conversion
def _binned_std(
    flat_idx: np.ndarray,
    n_bins: int,
    values: np.ndarray,
    bin_counts: np.ndarray | None = None,
    bin_mean: np.ndarray | None = None,
) -> np.ndarray:
    """
    Compute the standard deviation of values per bin.

    Parameters
    ----------
    flat_idx : array-like
        1D array of bin indices, same shape as values.
    n_bins: int
        number of bins
    values : array-like
        1D array of data values
    bin_counts : array-like, optional
        Precomputed counts per bin. If None, it will be computed using `_binned_count`.
    bin_mean : array-like, optional
        Precomputed mean per bin. If None, it will be computed using `_binned_mean`.

    Returns
    -------
    result : array-like
        1D array of length n_bins with the standard deviation per bin
    """
    if bin_counts is None:
        bin_counts = _binned_count(flat_idx, n_bins)

    if bin_mean is None:
        bin_mean = _binned_mean(flat_idx, n_bins, values, bin_counts)

    if np.iscomplexobj(values):
        # Use modulus for variance
        abs_values = np.abs(values)
        bin_sumsq = np.bincount(flat_idx, weights=abs_values**2, minlength=n_bins)
        bin_mean_sq = np.divide(
            bin_sumsq,
            bin_counts,
            out=np.full(n_bins, np.nan, dtype=bin_sumsq.dtype),
            where=bin_counts > 0,
        )
        abs_bin_mean = np.abs(bin_mean)
        variance = np.maximum(bin_mean_sq - abs_bin_mean**2, 0)
    else:
        bin_sumsq = np.bincount(flat_idx, weights=values**2, minlength=n_bins)
        bin_mean_sq = np.divide(
            bin_sumsq,
            bin_counts,
            out=np.full(n_bins, np.nan, dtype=bin_sumsq.dtype),
            where=bin_counts > 0,
        )
        variance = np.maximum(bin_mean_sq - bin_mean**2, 0)

    return np.sqrt(variance)


@handle_datetime_conversion
def _binned_min(flat_idx: np.ndarray, n_bins: int, values: np.ndarray) -> np.ndarray:
    """
    Compute the minimum of values per bin.

    Parameters
    ----------
    flat_idx : array-like
        1D array of bin indices, same shape as values.
    n_bins: int
        number of bins
    values : array-like
        1D array of data values

    Returns
    -------
    result : array-like
        1D array of length n_bins with the minimum per bin
    """
    if np.iscomplexobj(values):
        raise ValueError("Complex values are not supported for 'min' statistic.")

    output = np.full(n_bins, np.inf)
    np.minimum.at(output, flat_idx, values)
    output[output == np.inf] = np.nan
    return output


@handle_datetime_conversion
def _binned_max(flat_idx: np.ndarray, n_bins: int, values: np.ndarray) -> np.ndarray:
    """
    Compute the maximum of values per bin.

    Parameters
    ----------
    flat_idx : array-like
        1D array of bin indices, same shape as values.
    n_bins: int
        number of bins
    values : array-like
        1D array of data values

    Returns
    -------
    result : array-like
        1D array of length n_bins with the maximum per bin
    """
    if np.iscomplexobj(values):
        raise ValueError("Complex values are not supported for 'max' statistic.")

    output = np.full(n_bins, -np.inf)
    np.maximum.at(output, flat_idx, values)
    output[output == -np.inf] = np.nan
    return output


@handle_datetime_conversion
def _binned_apply_func(
    flat_idx: np.ndarray,
    n_bins: int,
    values: np.ndarray,
    func: Callable[[np.ndarray | list[np.ndarray]], float] = np.mean,
) -> np.ndarray:
    """
    Generic wrapper to apply any functions (e.g., percentile) to binned data.

    Parameters
    ----------
    flat_idx : array-like
        1D array of bin indices.
    n_bins : int
        Number of bins.
    values : array-like or list of array-like
        1D array (univariate) or list of 1D arrays (multivariate) of data values.
    func : Callable[[list[np.ndarray]], float]
        Function to apply to each bin. If multivariate, will receive a list of arrays.

    Returns
    -------
    result : np.ndarray
        1D array of length n_bins with results from func per bin.
    """
    sort_indices = np.argsort(flat_idx)
    sorted_flat_idx = flat_idx[sort_indices]

    # single or all variables can be passed as input values
    if is_multivariate := values.ndim == 2:
        sorted_values = [v[sort_indices] for v in values]
    else:
        sorted_values = [values[sort_indices]]

    unique_bins, bin_starts = np.unique(sorted_flat_idx, return_index=True)
    bin_ends = np.append(bin_starts[1:], len(sorted_flat_idx))

    result = np.full(n_bins, np.nan)
    for i, bin_idx in enumerate(unique_bins):
        if is_multivariate:
            bin_values = [v[bin_starts[i] : bin_ends[i]] for v in sorted_values]
        else:
            bin_values = sorted_values[0][bin_starts[i] : bin_ends[i]]
        result[bin_idx] = func(bin_values)

    return result



"""
This module provides functions to calculate various transfer function from wind stress to oceanic
velocity.
"""

import warnings
from typing import Callable, Optional

import numpy as np
from numpy import floating as _Floating
from numpy.lib.scimath import sqrt
from scipy.special import factorial, iv, kv  # type: ignore

from clouddrift.sphere import EARTH_DAY_SECONDS


def apply_transfer_function(
    x: np.ndarray,
    transfer_func: Callable[[np.ndarray], np.ndarray] | np.ndarray,
    dt: Optional[float] = None,
) -> np.ndarray:
    """
    Applies a transfer function in the frequency domain to an input time series.

    Parameters
    ----------
        x:  array_like
            Input one-dimensional time series.
        transfer_func: callable or numpy ndarray
            Transfer function to apply to the input `x`.
                - If a callable, it should accept a frequency array as input as shown in the examples below.
                - If a numpy ndarray, it must match the length of input `x` and represents the transfer function
            in the frequency domain, where the first half corresponds to positive frequencies and the
            second half corresponds to negative frequencies.
        dt: float, optional
            Time step size of the input time series, needed if `transfer_func` is callable, otherwise ignored.

    Returns
    -------
        y: numpy ndarray
            Output time series after applying the transfer function.

    Examples
    --------
        Apply a low-pass filter transfer function to a random time series:
        >>> import numpy as np
        >>> from clouddrift.transfer import apply_transfer_function
        >>> x = np.random.randn(100)
        >>> transfer_func = lambda omega: 1 / (1 + (omega / (2 * np.pi * 0.1))**2)  # Low-pass filter
        >>> dt = 1.0  # Time step needed for callable transfer functions
        >>> y = apply_transfer_function(x, transfer_func, dt)

        Directly provide as an array a transfer function which suppresses
        the negative frequencies of a complex time series:
        >>> import numpy as np
        >>> from clouddrift.transfer import apply_transfer_function
        >>> x = np.random.randn(100) + 1j * np.random.randn(100)
        >>> transfer_func = np.ones(100)
        >>> transfer_func[50:] = 0  # Suppress negative frequencies
        >>> y = apply_transfer_function(x, transfer_func)

        To a random wind stress, apply a slab wind transfer function at latitude 45 degrees,
        with a friction coefficient of 0.1 times the Coriolis frequency and a boundary layer
        depth of 50 meters:
        >>> import numpy as np
        >>> from clouddrift.sphere import coriolis_frequency, EARTH_DAY_SECONDS
        >>> from clouddrift.transfer import apply_transfer_function, slab_wind_transfer
        >>> tau = np.random.randn(100) + 1j * np.random.randn(100)  # Random wind stress
        >>> cor_freq = coriolis_frequency(45) * EARTH_DAY_SECONDS  # Coriolis frequency at 45 degrees latitude in radians per day
        >>> transfer_slab = lambda omega: slab_wind_transfer(omega, cor_freq[0] , friction=0.1 * cor_freq[0] / EARTH_DAY_SECONDS, bld=50)
        >>> dt = 1.0  # Time step needed for callable transfer functions
        >>> u = apply_transfer_function(tau, transfer_slab, dt)

    Raises
    ------
        ValueError
            If input `x` is not a one-dimensional array.
            If `transfer_func` is callable but `dt` is not provided.
            If the shape of `transfer_func` does not match the shape of `x`.
    """

    x = np.asarray(x)

    if x.ndim != 1:
        raise ValueError("Input x must be a 1D array.")
    n = x.size

    Xf = np.fft.fft(x)

    if callable(transfer_func):
        if dt is None:
            raise ValueError(
                "The `dt` argument must be provided when `transfer_func` is callable."
            )
        omega = 2 * np.pi * np.fft.fftfreq(n, dt)
        G = transfer_func(omega)
    else:
        if dt is not None:
            warnings.warn(
                "The `dt` argument is ignored when `transfer_func` is provided as an array.",
                UserWarning,
                stacklevel=2,
            )
        G = np.asarray(transfer_func)
        if G.shape != Xf.shape:
            raise ValueError(
                "Transfer function array must match the shape of input `x`."
            )

    Yf = G * Xf
    y = np.fft.ifft(Yf)

    return y


def _epanechnikov_kernel(window_size):
    # Epanechnikov kernel: k(x) = 0.75 * (1 - x^2) for |x| <= 1
    x = np.linspace(-1, 1, window_size + 2)[
        1:-1
    ]  # Exclude endpoints to avoid null weights
    kernel = 0.75 * (1 - x**2)
    return kernel


def apply_sliding_transfer_function(
    x: np.ndarray,
    transfer_func: Callable[[np.ndarray], np.ndarray] | np.ndarray,
    window_size: Optional[int] = None,
    step: int = 1,
    dt: Optional[float] = None,
    kernel: str = "uniform",
) -> tuple[np.ndarray, np.ndarray]:
    """
    Applies a transfer function in the frequency domain to a one-dimensional input time series in a
    sliding window fashion.
    This function returns the result for each window, as well as a weighted average of the results
    across all windows using a uniform (default) or Epanechnikov kernel.
    Mirror boundary conditions are used to extend the input time series so that the sliding window
    can be applied without losing data at the edges.

    Parameters
    ----------
        x:  array_like
            Input one-dimensional time series.
        transfer_func: callable or numpy ndarray
            Transfer function to apply to the input `x`.
                - If a callable, it should accept a frequency array as input as shown in the examples below.
                - If a numpy ndarray, it must match the length of input `x` and represents the transfer function
            in the frequency domain, where the first half corresponds to positive frequencies and the
            second half corresponds to negative frequencies.
        window_size: int, optional
            Size of the sliding window, needed if `transfer_func` is callable. It cannot be larger
            than the length of `x`.
        step: int
            Step size for the sliding window. Default is 1. Consecutive windows will not overlap if
            `step` is equal to or larger than `window_size` or the length of `transfer_func` if this
            one is a numpy array.
        dt: float, optional
            Time step size of the input time series, needed if `transfer_func` is callable, otherwise ignored.
        kernel: str, optional
            Type of kernel to use for weighting. Options are "uniform" (default), "epanechnikov",
            or any first letter of either. For non-overlapping windows, the chosen kernel is  irrelevant.

    Returns
    -------
        result: tuple of np.ndarray
            Tuple containing:
                - Array of results for each window (shape: (num_windows, window_size))
                - Weighted average result (shape: (n,))

    Examples
    --------
        Apply a low-pass filter in a sliding window:
        >>> import numpy as np
        >>> from clouddrift.transfer import apply_sliding_transfer_function
        >>> x = np.random.rand(100)
        >>> window_size = 10
        >>> step = 1
        >>> dt = 1.0
        >>> transfer_func = lambda omega: 1 / (1 + (omega / (2 * np.pi * 0.1))**2)  # Low-pass filter
        >>> result = apply_sliding_transfer_function(x, transfer_func, window_size, step, dt)
        The shape of the result is here (len(`x`) // `step` + 1, `window_size`) = (101, 10).

        Apply a downscaling function in non-overlapping windows:
        >>> import numpy as np
        >>> from clouddrift.transfer import apply_sliding_transfer_function
        >>> x = np.random.rand(100)
        >>> window_size = 10
        >>> step = 11
        >>> transfer_func = np.ones(window_size) / np.sqrt(2)
        >>> results, average_result = apply_sliding_transfer_function(x, transfer_func, window_size, step)
        The shape of results is here (len(`x`) // `step` + 1, `window_size`) = (91, 10).

        To a random wind stress hourly time series, apply an Ekman transfer function in a sliding window
        for an infinite boundary layer depth, an Ekman depth of 10 m, in order to obtain the Ekman
        velocity at a depth of 15 m, and use an Epanechnikov kernel for weighting:
        >>> import numpy as np
        >>> from clouddrift.sphere import coriolis_frequency, EARTH_DAY_SECONDS
        >>> from clouddrift.transfer import apply_transfer_function, slab_wind_transfer
        >>> tau = np.random.randn(240) + 1j * np.random.randn(240)  # Random wind stress
        >>> cor_freq = coriolis_frequency(45) * EARTH_DAY_SECONDS  # Coriolis frequency at 45 degrees latitude in radians per day
        >>> transfer_ekman = lambda omega: wind_transfer(omega, z=15, cor_freq= cor_freq,
        delta=10.0, mu=0.0, bld=np.inf, boundary_condition="no-slip")[0].squeeze()
        >>> dt = 1 / 24  # Time step needed for callable transfer functions
        >>> results, average_result = apply_sliding_transfer_function(tau, transfer_ekman, window_size=24, step=12, dt=dt, kernel="epanechnikov")
        The shape of results is here (len(`tau`) // `step` + 1, `window_size`) = (21, 24).

    Raises
    ------
        ValueError
            If `transfer_func` is a numpy array which is not 1D.
            If `transfer_func` is a numpy array which length does not match `window_size` if this one is provided.
            If `window_size` or `step` is not a positive integer.
            If `window_size` is greater than the length of `x`.
            If `step` is not a positive integer.
            If `dt` is not provided when `transfer_func` is callable.
            If `kernel` is not one of the allowed strings ("uniform", "epanechnikov", or their first letters).
        Warning
            If `dt` is provided when `transfer_func` is a numpy array, as it is ignored in this case.
    """
    x = np.asarray(x)
    n = x.size

    if callable(transfer_func):
        if window_size is None:
            raise ValueError(
                "`window_size` must be provided when transfer_func is callable."
            )
        if not isinstance(window_size, int) or window_size <= 0 or window_size > n:
            raise ValueError(
                "`window_size` must be a positive integer no greater than the length of `x`."
            )
        if not isinstance(step, int) or step <= 0:
            raise ValueError("`step` must be a positive integer.")
        if dt is None:
            raise ValueError(
                "`dt` (time step of input) must be provided when `transfer_func` is callable."
            )

    if not callable(transfer_func) and isinstance(transfer_func, np.ndarray):
        if transfer_func.ndim != 1:
            raise ValueError("`transfer_func` must be a 1D numpy array.")
        if window_size is None:
            window_size = transfer_func.shape[0]
        if window_size != transfer_func.shape[0]:
            raise ValueError(
                "When `transfer_func` is a numpy ndarray, its length must match `window_size` if this one"
                " is provided."
            )
        if not isinstance(step, int) or step <= 0:
            raise ValueError("`step` must be a positive integer.")
        if dt is not None:
            warnings.warn(
                "The `dt` argument is ignored when `transfer_func` is provided as an array.",
                stacklevel=2,
            )

    # Kernel selection (default is uniform)
    kernel_str = str(kernel).lower() if kernel is not None else "uniform"
    if window_size is None:
        raise ValueError("window_size must be specified and not None.")
    if kernel_str.startswith("e"):
        kernel_arr = _epanechnikov_kernel(window_size)
    elif kernel_str.startswith("u") or kernel_str == "":
        kernel_arr = np.ones(window_size)
    else:
        raise ValueError(
            "kernel must be 'uniform', 'epanechnikov', or their first letter."
        )

    # Extend x by half window_size at the beginning and end using mirror boundary condition
    before_window = window_size // 2
    after_window = window_size - before_window
    before_mirror = (
        x[1 : before_window + 1][::-1] if before_window > 0 else np.array([])
    )
    after_mirror = x[-after_window - 1 : -1][::-1] if after_window > 0 else np.array([])
    x_ = np.concatenate([before_mirror, x, after_mirror])

    n_ = (
        x_.size
    )  # length of extended array is n + before_window + after_window = n + window_size

    # initialize arrays for results
    result_sum = np.zeros(n_, dtype=np.complex128)
    result_weight = np.zeros(n_, dtype=np.float64)
    # result = []
    if window_size is None:
        raise ValueError("`window_size` must be specified and not None.")
    # number of windows is (n_ - window_size) // step + 1 = n // step + 1
    result = np.zeros((n // step + 1, window_size), dtype=np.complex128)

    # Sliding window application
    for start in range(0, n_ - window_size + 1, step):
        x_tmp = x_[start : start + window_size]
        if callable(transfer_func):
            result_tmp = apply_transfer_function(x_tmp, transfer_func, dt)
            result[start // step, :] = result_tmp
        else:
            if transfer_func.ndim == 1:
                result_tmp = apply_transfer_function(x_tmp, transfer_func)
                result[start // step, :] = result_tmp
            else:
                raise ValueError(
                    "`transfer_func` must be callable or a 1D numpy array."
                )
        weighted_result_tmp = result_tmp * kernel_arr
        result_sum[start : start + window_size] += weighted_result_tmp
        result_weight[start : start + window_size] += kernel_arr

    # average the results
    with np.errstate(divide="ignore", invalid="ignore"):
        average_result = result_sum / result_weight
    # truncate the average_result to the original length of x
    average_result = average_result[before_window : before_window + n]

    return result, average_result


def slab_wind_transfer(
    omega: float | np.ndarray,
    cor_freq: float,
    friction: float,
    bld: float,
    density: float = 1025.0,
) -> np.ndarray:
    """
    Compute the the transfer function in the case of a ocean slab layer.

    Parameters
    ----------
        omega: array_like or float
            Angular frequency of the wind stress forcing in radians per day.
        cor_freq: float
            Coriolis frequency, in radians per day.
        friction: float
            Friction coefficient, in s-1.
        bld: float
            Thickness of the slab layer, in meters.
        density: float
            Seawater density, in kg m-3. Default is 1025.

    Returns
    -------
        G: np.ndarray
            The transfer function from wind stress to oceanic velocity in units of kg-1 m 2 s.

    References
    ----------
    Pollard, R. T., and R. C. Millard Jr., 1970: Comparison between
    observed and simulated wind-generated inertial oscillations.
    Deep-Sea Res., 17, 813–816, https://doi.org/10.1016/0011-7471(70) 90043-4.
    """
    # check that the boundary layer depth is positive
    if bld < 0:
        raise ValueError("Boundary layer depth bld must be positive.")

    # check that the density is positive
    if density < 0:
        raise ValueError("Density density must be positive.")

    # omega can be scalars or arrays, convert to arrays here
    omega_ = np.atleast_1d(omega)

    # convert to radians per second
    omega_ = omega_ / EARTH_DAY_SECONDS
    cor_freq_ = cor_freq / EARTH_DAY_SECONDS

    G = 1 / (density * bld * (friction + 1j * (omega_ + cor_freq_)))

    return G


def wind_transfer(
    omega: float | np.ndarray,
    z: float | np.ndarray,
    cor_freq: float,
    delta: float,
    mu: float,
    bld: float,
    boundary_condition: str = "no-slip",
    method: str = "lilly",
    density: float = 1025.0,
) -> tuple[
    float | _Floating | np.ndarray,
    float | _Floating | np.ndarray,
    float | _Floating | np.ndarray,
]:
    """
    Compute the transfer function from wind stress to oceanic velocity based on the physically-based
    models of Elipot and Gille (2009) and Lilly and Elipot (2021).

    Parameters
    ----------
        omega: array_like or float
            Angular frequency of the wind stress forcing in radians per day.
        z: array_like or float
            Depth in the ocean, positive downwards, in meters.
        cor_freq: float
            Coriolis frequency, in radians per day.
        delta: float
            Ekman depth, in meters.
        mu: float
            Madsen depth, in meters.
        bld: float
            Boundary layer depth, in meters.
        boundary_condition: str
            Bottom boundary condition at the base of the ocean surface boundary layer.
            Options are "no-slip" (Default) or "free-slip".
        method: str
            Method to compute the transfer function. Options are "lilly" (Default and preferred method)
            or "elipot".
        density: float
            Seawater density, in kg m-3. Default is 1025.

    Returns
    -------
        G: np.ndarray
            The transfer function from wind stress to oceanic velocity in units of kg-1 m 2 s.
        ddelta: np.ndarray
            The gradient of transfer function from the rate of change of the Ekman depth.
        dh: np.ndarray
            The gradient of transfer function from to the rate of change of the boundary layer depth.

    Examples
    --------
        To calculate the transfer function of Lilly and Elipot (2021) corresponding to the Ekman case
         of infinite depth ocean surface boundary layer and constant vertical eddy viscosity, at the zero frequency:

         >>> omega = 0
         >>> z = np.linspace(0, 100, 100)
         >>> cor_freq = 2 * np.pi * 1
         >>> K0 = 1e-4
         >>> delta = np.sqrt(2 * K0 / cor_freq / 86400)
         >>> mu = 0
         >>> G = wind_transfer(omega, z, cor_freq, delta, mu, np.inf, "no-slip")

    Raises
    ------
    ValueError
        If the depth z is not positive.

    References
    ----------
    [1] Elipot, S., and S. T. Gille (2009). Ekman layers in the Southern Ocean: spectral models and
    observations, vertical viscosity and boundary layer depth Ocean Sci., 5, 115–139, 2009, doi:10.5194/os-5-115-2009.

    [2] Lilly, J. M. and S. Elipot (2021). A unifying perspective on transfer function solutions to
    the unsteady Ekman problem. Fluids, 6 (2): 85, 1--36. doi:10.3390/fluids6020085.

    """
    # check that z is positive
    if np.any(z < 0):
        raise ValueError("Depth z must be positive.")

    # check that the boundary layer depth is positive
    if bld < 0:
        raise ValueError("Boundary layer depth bld must be positive.")

    # check that the Ekman depth is positive
    if delta < 0:
        raise ValueError("Ekman depth delta must be positive.")

    # check that the Madsen depth is positive
    if mu < 0:
        raise ValueError("Madsen depth mu must be positive.")

    # check that the density is positive
    if density < 0:
        raise ValueError("Density density must be positive.")

    # check that the boundary condition is valid
    if boundary_condition not in ["no-slip", "free-slip"]:
        raise ValueError("Boundary condition must be 'no-slip' or 'free-slip'.")

    # check that the method is valid
    if method not in ["lilly", "elipot"]:
        raise ValueError("Method must be 'lilly' or 'elipot'.")

    # z and omega can be scalars or arrays, convert to arrays here
    z_ = np.atleast_1d(z)
    omega_ = np.atleast_1d(omega)

    # convert to radians per second
    omega_ = omega_ / EARTH_DAY_SECONDS
    cor_freq_ = cor_freq / EARTH_DAY_SECONDS

    # Create the gridmesh of frequency and depth
    [omega_grid, z_grid] = np.meshgrid(omega_, z_)

    # apply the selected method and boundary condition, see references [1] and [2].
    if boundary_condition == "no-slip":
        if method == "lilly":
            G = _wind_transfer_no_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )
        else:
            G = _wind_transfer_elipot_no_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )
    elif boundary_condition == "free-slip":
        if method == "lilly":
            G = _wind_transfer_free_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )
        else:
            G = _wind_transfer_elipot_free_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )

    # set G to nan where z > bld; may be mathematcially valid but not physically meaningful
    G[z_grid > bld] = np.nan

    # analytical gradients of the transfer function for mu = 0 and no-slip, see reference [2]
    if boundary_condition == "no-slip" and method == "lilly" and mu == 0:
        with np.errstate(divide="ignore", invalid="ignore"):
            s = np.sign(cor_freq_) * np.sign(1 + omega_grid / cor_freq_)
            Gamma = (
                sqrt(2) * _rot(s * np.pi / 4) * sqrt(np.abs(1 + omega_grid / cor_freq_))
            )
            ddelta1 = (
                (Gamma * (bld / delta) * np.tanh(Gamma * (bld / delta)) - 1) * G / delta
            )

            numer = np.exp(Gamma * (-z_grid / delta)) + np.exp(
                -Gamma * (2 * bld - z_grid) / delta
            )
            denom = 1 + np.exp(-Gamma * (2 * bld / delta))
            ddelta2 = (
                -2
                / (delta**2 * np.abs(cor_freq_) * density)
                * (bld - z_grid)
                / delta
                * (numer / denom)
            )

            dG_ddelta = ddelta1 + ddelta2

            dbld1 = -Gamma / delta * np.tanh(Gamma * (bld / delta)) * G
            dbld2 = 2 / (delta**2 * np.abs(cor_freq_) * density) * (numer / denom)
            dG_dbld = dbld1 + dbld2

    else:
        dG_ddelta = np.array([], dtype=float)  # Empty array for consistency
        dG_dbld = np.array([], dtype=float)

    return G, dG_ddelta, dG_dbld


def _wind_transfer_no_slip(
    omega_grid: np.ndarray,
    z_grid: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary condition,
    Lilly's method, see reference [2].
    """

    zo = delta**2 / mu if mu != 0 else np.inf
    s = np.sign(coriolis_frequency) * np.sign(1 + omega_grid / coriolis_frequency)

    xiz, xih, xi0 = _xis(s, zo, delta, z_grid, omega_grid, coriolis_frequency, bld)

    with np.errstate(divide="ignore", invalid="ignore"):
        if bld == np.inf:
            if mu == 0:
                # Ekman solution
                coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
                    delta * np.abs(coriolis_frequency) * density
                )
                G = (
                    coeff
                    * (
                        np.exp(
                            -(1 + s * 1j)
                            * (z_grid / delta)
                            * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
                        )
                    )
                    / (sqrt(np.abs(1 + omega_grid / coriolis_frequency)))
                )
            elif delta == 0:
                # Madsen solution
                coeff = 4 / (density * np.abs(coriolis_frequency) * mu)
                G = coeff * kv(
                    0,
                    2
                    * sqrt(2)
                    * _rot(s * np.pi / 4)
                    * sqrt((z_grid / mu) * np.abs(1 + omega_grid / coriolis_frequency)),
                )
            else:
                # mixed solution
                k0z = kv(0, xiz)
                k10 = kv(1, xi0)
                coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
                    delta
                    * np.abs(coriolis_frequency)
                    * density
                    * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
                )
                G = coeff * k0z / k10
        else:
            if mu == 0:
                # finite layer Ekman
                coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
                    delta
                    * np.abs(coriolis_frequency)
                    * density
                    * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
                )
                argh = (
                    sqrt(2)
                    * _rot(s * np.pi / 4)
                    * (bld / delta)
                    * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
                )
                argz = (
                    sqrt(2)
                    * _rot(s * np.pi / 4)
                    * (z_grid / delta)
                    * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
                )

                numer = np.exp(-argz) - np.exp(argz) * np.exp(-2 * argh)
                denom = 1 + np.exp(-2 * argh)
                G = coeff * np.divide(numer, denom)

                # solution at inertial frequency
                bool_idx = omega_grid == -coriolis_frequency
                G[bool_idx] = 2 / (
                    (density * np.abs(coriolis_frequency) * delta**2)
                    * (bld - z_grid[bool_idx])
                )

            elif delta == 0:
                # finite layer Madsen
                coeff = 4 / (density * np.abs(coriolis_frequency) * mu)
                argz = (
                    2
                    * sqrt(2)
                    * _rot(s * np.pi / 4)
                    * sqrt((z_grid / mu) * np.abs(1 + omega_grid / coriolis_frequency))
                )
                argh = (
                    2
                    * sqrt(2)
                    * _rot(s * np.pi / 4)
                    * sqrt((bld / mu) * np.abs(1 + omega_grid / coriolis_frequency))
                )
                k0z, i0z, k0h, i0h, _, _ = _bessels_noslip(argz, argh)
                G = coeff * (k0z - i0z * k0h / i0h)

                # solution at inertial frequency
                bool_idx = omega_grid == -coriolis_frequency
                if isinstance(z_grid, np.ndarray):
                    G[bool_idx] = 0.5 * coeff * np.log(bld / z_grid[bool_idx])
                else:
                    G = 0.5 * coeff * np.log(bld / z_grid)
            else:
                # finite layer mixed
                G = _wind_transfer_general_no_slip(
                    omega_grid, z_grid, coriolis_frequency, delta, mu, bld, density
                )

    return G


def _wind_transfer_general_no_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary condition,
    general form, Lilly's method, see reference [2].
    """
    zo = delta**2 / mu if mu != 0 else np.inf
    s = np.sign(coriolis_frequency) * np.sign(1 + omega / coriolis_frequency)
    xiz, xih, xi0 = _xis(s, zo, delta, z, omega, coriolis_frequency, bld)

    with np.errstate(divide="ignore", invalid="ignore"):
        coeff = (
            sqrt(2)
            * _rot(-s * np.pi / 4)
            / (
                delta
                * density
                * np.abs(coriolis_frequency)
                * sqrt(np.abs(1 + omega / coriolis_frequency))
            )
        )

        k0z, i0z, k0h, i0h, k10, i10 = _bessels_noslip(xiz, xih, xi0=xi0)
        numer = k0z / k10 - (k0h / k10) * (i0z / i0h)
        denom = 1 + (k0h / k10) * (i10 / i0h)
        G = coeff * np.divide(numer, denom)

        # large argument approximation
        bool_idx = np.log10(np.abs(xiz)) > 2.9
        G[bool_idx] = _wind_transfer_general_no_slip_expansion(
            omega[bool_idx], z[bool_idx], coriolis_frequency, delta, mu, bld, density
        )

    # inertial limit
    G = _wind_transfer_inertiallimit(
        G, omega, z, coriolis_frequency, delta, mu, bld, density
    )

    return G


def _wind_transfer_general_no_slip_expansion(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Compute the transfer function from wind stress to oceanic velocity with no-slip boundary,
    large argument approximation, Lilly's method, see reference [2].
    """
    zo = delta**2 / mu if mu != 0 else np.inf
    s = np.sign(coriolis_frequency) * np.sign(1 + omega / coriolis_frequency)
    xiz, xih, xi0 = _xis(s, zo, delta, z, omega, coriolis_frequency, bld)
    coeff = (
        sqrt(2)
        * _rot(-s * np.pi / 4)
        / (
            delta
            * density
            * np.abs(coriolis_frequency)
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
    )
    k0z, i0z, k0h, i0h, k10, i10 = _besseltildes_noslip(xiz, xih, xi0, 30)

    numer = np.exp(xi0 - xiz) * i0h * k0z - np.exp(xi0 + xiz - 2 * xih) * k0h * i0z
    denom = i0h * k10 + np.exp(2 * xi0 - 2 * xih) * k0h * i10
    G = coeff * np.divide(numer, denom)

    bool_idx = omega == -coriolis_frequency
    G[bool_idx] = (
        4
        * zo
        / (
            density
            * np.abs(coriolis_frequency)
            * delta**2
            * np.divide(sqrt(1 + bld / zo) - sqrt(1 + z / zo), (1 + z / zo) ** 0.25)
        )
    )

    return G


def _wind_transfer_inertiallimit(
    G: np.ndarray,
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_freq: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary, inertial limit,
    Lilly's method, see reference [2].
    """
    zo = delta**2 / mu if mu != 0 else np.inf
    bool_idx = omega == -coriolis_freq

    if np.any(bool_idx):
        if not np.isinf(bld) and not np.isinf(zo):
            G[bool_idx] = (2 / (density * np.abs(coriolis_freq) * mu)) * np.log(
                (1 + bld / zo) / (1 + z[bool_idx] / zo)
            )
        elif not np.isinf(bld) and np.isinf(zo):
            G[bool_idx] = (2 / (density * np.abs(coriolis_freq) * delta**2)) * (
                bld - z[bool_idx]
            )
        else:
            G[bool_idx] = np.inf

    return G


def _wind_transfer_free_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with free-slip boundary condition,
    Lilly's method, see reference [2].
    """

    zo = delta**2 / mu if mu != 0 else np.inf

    s = np.sign(coriolis_frequency) * np.sign(1 + omega / coriolis_frequency)

    xiz, xih, xi0 = _xis(s, zo, delta, z, omega, coriolis_frequency, bld)

    with np.errstate(divide="ignore", invalid="ignore"):
        if delta != 0.0 and mu != 0.0:
            coeff = (
                sqrt(2)
                * _rot(-s * np.pi / 4)
                / (
                    delta
                    * density
                    * np.abs(coriolis_frequency)
                    * sqrt(np.abs(1 + omega / coriolis_frequency))
                )
            )
            k0z, i0z, k1h, i1h, k10, i10 = _bessels_freeslip(xiz, xih, xi0=xi0)
            numer = i0z * k1h + i1h * k0z
            denom = i1h * k10 - i10 * k1h
            G = coeff * np.divide(numer, denom)

        elif mu == 0.0:
            coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
                delta
                * np.abs(coriolis_frequency)
                * density
                * sqrt(np.abs(1 + omega / coriolis_frequency))
            )
            cosharg = (
                sqrt(2)
                * _rot(s * np.pi / 4)
                * (bld - z)
                / delta
                * sqrt(np.abs(1 + omega / coriolis_frequency))
            )
            sinharg = (
                sqrt(2)
                * _rot(s * np.pi / 4)
                * bld
                / delta
                * sqrt(np.abs(1 + omega / coriolis_frequency))
            )
            G = coeff * np.cosh(cosharg) / np.sinh(sinharg)

        elif delta == 0.0:
            k0z, i0z, k1h, i1h, _, _ = _bessels_freeslip(xiz, xih)

            K1 = 0.5 * mu * np.abs(coriolis_frequency)

            coeff = 2 / (density * K1)

            G = coeff * (k0z + k1h * i0z / i1h)

    return G


def _wind_transfer_elipot_no_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary condition,
    Elipot's method, see reference [1].
    """

    zo = delta**2 / mu if mu != 0 else np.inf
    K0 = 0.5 * delta**2 * np.abs(coriolis_frequency)
    K1 = 0.5 * mu * np.abs(coriolis_frequency)

    with np.errstate(divide="ignore", invalid="ignore"):
        delta1 = sqrt(2 * K0 / (omega + coriolis_frequency))
        delta2 = K1 / (omega + coriolis_frequency)
        xiz = 2 * sqrt(1j * (zo + z) / delta2)
        xi0 = 2 * sqrt(1j * zo / delta2)
        xih = 2 * sqrt(1j * (zo + bld) / delta2)

        if bld == np.inf:
            if K1 == 0:
                # Ekman solution
                coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
                G = coeff * np.exp(-z * (1 + 1j) / delta1)
            elif K0 == 0:
                # Madsen solution
                coeff = 2 / (density * K1)
                k0z = kv(0, 2 * sqrt(1j * z / delta2))
                G = coeff * k0z
            else:
                # Mixed solution
                coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
                G = coeff * kv(0, xiz) / kv(1, xi0)
        else:
            if K1 == 0:
                # Finite-layer Ekman
                coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
                numer = np.sinh((1 + 1j) * (bld - z) / delta1)
                denom = np.cosh((1 + 1j) * bld / delta1)
                G = coeff * numer / denom
            elif K0 == 0:
                # Finite-layer Madsen
                coeff = 2 / (density * K1)
                k0z = kv(0, 2 * sqrt(1j * z / delta2))
                k0h = kv(0, 2 * sqrt(1j * bld / delta2))
                i0z = iv(0, 2 * sqrt(1j * z / delta2))
                i0h = iv(0, 2 * sqrt(1j * bld / delta2))
                G = coeff * (k0z - (k0h * i0z / i0h))
            else:
                # Finite-layer mixed solution
                coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
                k0z, i0z, k0h, i0h, k10, i10 = _bessels_noslip(xiz, xih, xi0)
                numer = i0h * k0z - k0h * i0z
                denom = i10 * k0h + k10 * i0h
                G = coeff * numer / denom

    return G


def _wind_transfer_elipot_free_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with free-slip boundary condition,
    Elipot's method, see reference [1].
    """

    zo = delta**2 / mu if mu != 0 else np.inf
    K0 = 0.5 * delta**2 * np.abs(coriolis_frequency)
    K1 = 0.5 * mu * np.abs(coriolis_frequency)

    with np.errstate(divide="ignore", invalid="ignore"):
        delta1 = sqrt(2 * K0 / (omega + coriolis_frequency))
        delta2 = K1 / (omega + coriolis_frequency)
        xiz = 2 * sqrt(1j * (zo + z) / delta2)
        xi0 = 2 * sqrt(1j * zo / delta2)
        xih = 2 * sqrt(1j * (zo + bld) / delta2)

        if K1 == 0.0:
            # Finite-layer Ekman
            coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
            numer = np.cosh((1 + 1j) * (bld - z) / delta1)
            denom = np.sinh((1 + 1j) * bld / delta1)
            G = coeff * numer / denom
        elif K0 == 0.0:
            # Finite-layer Madsen
            coeff = 2 / (density * K1)
            k0z = kv(0, 2 * sqrt(1j * z / delta2))
            k0h = kv(0, 2 * sqrt(1j * bld / delta2))
            i0z = iv(0, 2 * sqrt(1j * z / delta2))
            i1h = iv(1, 2 * sqrt(1j * bld / delta2))
            G = coeff * (k0z - (k0h * i0z / i1h))
        else:
            # Finite-layer mixed solution
            coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
            k1h = kv(1, xih)
            i0z = iv(0, xiz)
            i1h = iv(1, xih)
            k10 = kv(1, xi0)
            i10 = iv(1, xi0)
            k0z = kv(0, xiz)
            numer = k1h * i0z + i1h * k0z
            denom = i1h * k10 - i10 * k1h
            G = coeff * numer / denom

    return G


def _bessels_freeslip(
    xiz: float | np.ndarray,
    xih: float | np.ndarray,
    xi0: float | np.ndarray | None = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute the Bessel functions for the free-slip boundary condition for the xsi(z),
    xsi(h), and xsi(0) functions.
    """
    # Convert inputs to numpy arrays
    xiz = np.asarray(xiz)
    xih = np.asarray(xih)

    # Ensure all outputs are numpy arrays
    k0z = np.asarray(kv(0, xiz))
    i0z = np.asarray(iv(0, xiz))
    k1h = np.asarray(kv(1, xih))
    i1h = np.asarray(iv(0, xih))

    if xi0 is not None:
        xi0 = np.asarray(xi0)
        k10 = np.asarray(kv(1, xi0))
        i10 = np.asarray(iv(1, xi0))
    else:
        # Create nan values as numpy arrays with same shape as k0z
        k10 = np.full_like(k0z, np.nan)
        i10 = np.full_like(k0z, np.nan)

    return k0z, i0z, k1h, i1h, k10, i10


def _bessels_noslip(
    xiz: float | np.ndarray,
    xih: float | np.ndarray,
    xi0: float | np.ndarray | None = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute the Bessel functions for the no-slip boundary condition for the xsi(z),
    xsi(h), and xsi(0) functions.
    """
    # Convert inputs to numpy arrays
    xiz = np.asarray(xiz)
    xih = np.asarray(xih)

    # Ensure all outputs are numpy arrays
    k0z = np.asarray(kv(0, xiz))
    i0z = np.asarray(iv(0, xiz))
    k0h = np.asarray(kv(0, xih))
    i0h = np.asarray(iv(0, xih))

    if xi0 is not None:
        xi0 = np.asarray(xi0)
        k10 = np.asarray(kv(1, xi0))
        i10 = np.asarray(iv(1, xi0))
    else:
        # Create nan values as numpy arrays with same shape as k0z
        k10 = np.full_like(k0z, np.nan)
        i10 = np.full_like(k0z, np.nan)

    return k0z, i0z, k0h, i0h, k10, i10


def _besseltildes_noslip(
    xiz: float | np.ndarray,
    xih: float | np.ndarray,
    xi0: float | np.ndarray,
    nterms: int = 30,
) -> tuple[
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
]:
    """
    Compute the n-term expansion about the large-argument exponential behavior of Bessel functions
    for the xsi(z), xsi(h), and xsi(0) functions.
    """
    k0z = kvtilde(0, xiz, nterms)
    i0z = ivtilde(0, xiz, nterms)
    k0h = kvtilde(0, xih, nterms)
    i0h = ivtilde(0, xih, nterms)
    k10 = kvtilde(1, xi0, nterms)
    i10 = ivtilde(1, xi0, nterms)

    return k0z, i0z, k0h, i0h, k10, i10


def kvtilde(
    nu: int,
    z: float | np.ndarray,
    nterms: int = 30,
) -> float | np.ndarray:
    """
    Compute the n-term expansion about the large-argument exponential behavior of the modified Bessel
    function of the second kind of real order (kv).
    """
    z = np.asarray(
        z, dtype=np.complex128
    )  # Ensure z is an array for vectorized operations
    sizez = z.shape
    z = z.ravel()

    # Prepare zk for vectorized computation
    zk = np.tile(z, (nterms, 1)).T
    zk[:, 0] = 1.0
    zk = np.cumprod(zk, axis=1)

    k = np.arange(nterms)
    ak = 4.0 * nu**2 - (2.0 * k - 1.0) ** 2
    ak[0] = 1.0
    ak = np.cumprod(ak) / (factorial(k) * (8.0**k))

    # Handling potential non-finite values in ak
    if not np.all(np.isfinite(ak)):
        first_nonfinite = np.where(~np.isfinite(ak))[0][0]
        ak = ak[:first_nonfinite]
        zk = zk[:, :first_nonfinite]

    K = sqrt(np.pi / (2.0 * z)) * (np.dot(1.0 / zk, ak))
    K = K.reshape(sizez)

    return K


def ivtilde(
    nu: int,
    z: float | np.ndarray,
    nterms: int = 30,
) -> float | np.ndarray:
    """
    Compute the n-term expansion about the large-argument exponential behavior of the
    Modified Bessel function of the first kind of real order (iv).
    """
    z = np.asarray(
        z, dtype=np.complex128
    )  # Ensure z is an array for vectorized operations
    sizez = z.shape
    z = z.ravel()

    # Prepare zk for vectorized computation with alternating signs for each term
    zk = np.tile(-z, (nterms, 1)).T
    zk[:, 0] = 1.0
    zk = np.cumprod(zk, axis=1)

    k = np.arange(nterms)
    ak = 4.0 * nu**2 - (2.0 * k - 1.0) ** 2
    ak[0] = 1.0
    ak = np.cumprod(ak) / (factorial(k) * (8.0**k))

    # Handling potential non-finite values in ak
    if not np.all(np.isfinite(ak)):
        first_nonfinite = np.where(~np.isfinite(ak))[0][0]
        ak = ak[:first_nonfinite]
        zk = zk[:, :first_nonfinite]

    I = 1.0 / sqrt(2.0 * z * np.pi) * (np.dot(1.0 / zk, ak))
    I = I.reshape(sizez)

    return I


def _xis(
    s: float,
    zo: float,
    delta: float,
    z: float | np.ndarray,
    omega: float | np.ndarray,
    coriolis_frequency: float,
    bld: float,
) -> tuple[float | np.ndarray, float | np.ndarray, float | np.ndarray]:
    """
    Compute the complex-valued "xsi" functions.
    """
    with np.errstate(divide="ignore", invalid="ignore"):
        xiz = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(z, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        xiz = np.where(np.isfinite(xiz), xiz, np.nan)
        xih = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(bld, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        xih = np.where(np.isfinite(xih), xih, np.nan)
        xi0 = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
        xi0 = np.where(np.isfinite(xi0), xi0, np.nan)

    return xiz, xih, xi0


def _rot(x: float) -> complex:
    """
    Compute the complex-valued rotation.
    """
    return np.exp(1j * x)
